#!/usr/bin/env python

import argparse
import sys
from io import BytesIO
from definitions import ShellcodeDefinitions 
from transformer import ShellcodeTransformer


def main(args):

    raw_shellcode:BytesIO = args.infile.read()
    definition = ShellcodeDefinitions.get_definition(args.format)
    output = ShellcodeTransformer(definition).transform_shellcode(raw_shellcode)
    print(output)


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument('infile', nargs='?', type=argparse.FileType('rb'), default=sys.stdin.buffer)
    parser.add_argument('-f', help='Output format', choices=ShellcodeDefinitions.list_definitions(), required=True, dest='format')
    parser.add_argument('-o', '--outfile', help='Output file')
    args = parser.parse_args()

    main(args)


import ShellcodeFormatter.definitions as ShellcodeDefinitions
import ShellcodeFormatter.transformer as ShellcodeTransformer

ShellcodeDefinitions = ShellcodeDefinitions.ShellcodeDefinitions
ShellcodeTransformer = ShellcodeTransformer.ShellcodeTransformer


class ShellcodeDefinition:

    def __init__(self, name, output_format, byte_format, byte_separator, bytes_per_line:int, line_format, first_line_format=None, last_line_format=None):
        self.name = name
        self.output_format = output_format
        self.byte_format = byte_format
        self.byte_separator = byte_separator
        self.bytes_per_line = bytes_per_line
        self.line_format = line_format
        self.first_line_format = first_line_format
        self.last_line_format = last_line_format

class ShellcodeDefinitions:

    def get_definition(name):
        return definitions[name]
    
    def list_definitions():
        return definitions.keys()


definitions = {
    "csharp": ShellcodeDefinition(
        name="csharp",
        output_format="byte[] buf = new byte[ ~~BYTE_COUNT~~ ] { \n~~LINES~~};",
        byte_format="0xff",
        byte_separator=",",
        bytes_per_line=16,
        line_format="~~BYTES~~,\n",
        last_line_format="~~BYTES~~\n"
    ),
    "powershell": ShellcodeDefinition(
        name="powershell",
        output_format="[Byte[]] $buf = ~~LINES~~",
        byte_format="0xff",
        byte_separator=",",
        bytes_per_line=0,
        line_format="~~BYTES~~"
    ),
    "python": ShellcodeDefinition(
        name="python",
        output_format="buf = b''\n~~LINES~~",
        byte_format="\\xff",
        byte_separator="",
        bytes_per_line=16,
        line_format="buf += b'~~BYTES~~'\n"
    ),
    "vba": ShellcodeDefinition(
        name="vba",
        output_format="Dim buf as Variant\n~~LINES~~",
        byte_format="255",
        byte_separator=", ",
        bytes_per_line=16,
        first_line_format="buf = Array(~~BYTES~~)\n",
        line_format='buf = Split(Join(buf, ",") & "," & Join(Array(~~BYTES~~), ","), ",")\n'
    ),
    "java": ShellcodeDefinition(
        name="java",
        output_format="byte[] buf = new byte[]\n{\n~~LINES~~\n};",
        byte_format="(byte) 0xff",
        byte_separator=",",
        bytes_per_line=8,
        line_format="\t~~BYTES~~,\n",
        last_line_format="\t~~BYTES~~"
    ),
    "c": ShellcodeDefinition(
        name="c",
        output_format="unsigned char buf[] = \n~~LINES~~;",
        byte_format="\\xff",
        byte_separator="",
        bytes_per_line=16,
        line_format="\"~~BYTES~~\"\n",
        last_line_format="\"~~BYTES~~\""
    ),
    "bash": ShellcodeDefinition(
        name="bash",
        output_format="export buf=\\\n~~LINES~~",
        byte_format="\\xff",
        byte_separator="",
        bytes_per_line=16,
        line_format="$'~~BYTES~~'\\\n",
        last_line_format="$'~~BYTES~~'"
    ),
    "ruby": ShellcodeDefinition(
        name="ruby",
        output_format="buf = \n~~LINES~~",
        byte_format="\\xff",
        byte_separator="",
        bytes_per_line=16,
        line_format="\"~~BYTES~~\" +\n",
        last_line_format="\"~~BYTES~~\"\n"
    ),
    "perl": ShellcodeDefinition(
        name="perl",
        output_format="my $buf = \n~~LINES~~",
        byte_format="\\xff",
        byte_separator="",
        bytes_per_line=16,
        line_format="\"~~BYTES~~\" .\n",
        last_line_format="\"~~BYTES~~\"\n"
    ),
    "go": ShellcodeDefinition(
        name="go",
        output_format="buf := []byte{\n~~LINES~~}",
        byte_format="0xff",
        byte_separator=",",
        bytes_per_line=16,
        line_format="\t~~BYTES~~,\n",
        last_line_format="\t~~BYTES~~"
    ),
    "rust": ShellcodeDefinition(
        name="rust",
        output_format="let buf = [\n~~LINES~~];",
        byte_format="0xff",
        byte_separator=",",
        bytes_per_line=16,
        line_format="\t~~BYTES~~,\n",
        last_line_format="\t~~BYTES~~"
    ),
}


from io import BytesIO
from ShellcodeFormatter.definitions import ShellcodeDefinition


class ShellcodeTransformer:

    def __init__(self, definition:ShellcodeDefinition):
        self.definition = definition

    def transform_shellcode(self, shellcode:BytesIO) -> str:
        return transform_shellcode(shellcode, self.definition)


def _split_lines(bytes, definition:ShellcodeDefinition) -> list:
    """
    Splits the raw bytes of the shellcode into a list of lines to be processed
    against the definition
    """

    if definition.bytes_per_line == 0:
        line_length = len(bytes)
    else:
        line_length = definition.bytes_per_line
    
    raw_lines:list = [bytes[i:i+line_length] for i in range(0, len(bytes), line_length)]

    return raw_lines

def _format_lines(raw_lines:list, definition:ShellcodeDefinition) -> str:
    """
    Breaks each line down into a list of bytes, then formats the bytes into a string
    based on the definition. then joins the bytes to match the line format.

    Supported byte formats:
    - ff (hex lowercase)
    - FF (hex uppercase)
    - 255 (decimal)
    - 377 (octal)

    Any prefix or suffix will be left untouched, only the above characters will be replaced


    Returns a string of the formatted lines
    """
    lines = list();

    for line_number, raw_line in enumerate(raw_lines):

        if "ff" in definition.byte_format:
            # Lowercase hex format

            bytes = []
            for raw_byte in raw_line:
                # raw_byte = b'\x01'
                byte = definition.byte_format.replace("ff", f"{raw_byte:02x}")
                bytes.append(byte)
                
            line = definition.byte_separator.join(bytes)


        elif "FF" in definition.byte_format:
            # Uppercase hex format
            bytes = []
            for raw_byte in raw_line:
                # raw_byte = b'\x01'
                byte = definition.byte_format.replace("FF", f"{raw_byte:02X}")
                bytes.append(byte)
                
            line = definition.byte_separator.join(bytes)

        elif "255" in definition.byte_format:
            # Decimal format

            bytes = []
            for raw_byte in raw_line:
                # raw_byte = b'\x01'
                byte = definition.byte_format.replace("255", f"{raw_byte:03d}")
                bytes.append(byte)
                
            line = definition.byte_separator.join(bytes)

        elif "377" in definition.byte_format:
            # Octal format

            bytes = []
            for raw_byte in raw_line:
                # raw_byte = b'\x01'
                byte = definition.byte_format.replace("377", f"{raw_byte:03o}")
                bytes.append(byte)
                
            line = definition.byte_separator.join(bytes)

        else:
            raise Exception("Invalid byte format")


        if line_number == 0 and definition.first_line_format is not None:
            # First line
            lines.append(definition.first_line_format.replace("~~BYTES~~", line))
        elif line_number == len(raw_lines) - 1 and definition.last_line_format is not None:
            # Last line
            lines.append(definition.last_line_format.replace("~~BYTES~~", line))
        else:
            # All other lines
            lines.append(definition.line_format.replace("~~BYTES~~", line))

    # Combne the lines into a single string
    lines = "".join(lines)
    return lines


def transform_shellcode(shellcode:BytesIO, definition:ShellcodeDefinition) -> str:
    """
    Transforms a raw binary file into a string based on the provided definition
    """

    raw_lines = _split_lines(shellcode, definition)

    lines = _format_lines(raw_lines, definition)
    
    # Replace the output format with the lines
    output = definition.output_format.replace("~~LINES~~", lines)
    # Add in the byte count
    output = output.replace("~~BYTE_COUNT~~", str(len(shellcode)))

    return output


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
     name='ShellcodeFormatter',
     version='1.0.1',
     author="Eth0",
     author_email="e@eth0.re",
     description="Transform binary shellcode into various formats",
     long_description=long_description,
   long_description_content_type="text/markdown",
     url="https://github.com/eth0-re/ShellcodeFormatter",
     download_url="https://github.com/eth0-re/ShellcodeFormatter/archive/refs/tags/v1.0.1.tar.gz",
     packages=['ShellcodeFormatter'],
     classifiers=[
         "Programming Language :: Python :: 3",
         "License :: OSI Approved :: MIT License",
         "Operating System :: OS Independent",
     ],
 )